import argparse
import datetime
import getpass
import logging
import os
import sys

from lightsteem.client import Client
from lightsteem.datastructures import Operation
from pymongo import MongoClient

from .utils import generate_keys

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.basicConfig()


class GiftCodeManager:

    def __init__(self, mongo_connection_string):
        self.client = MongoClient(mongo_connection_string)
        self.db = self.client.infestor
        self.table = self.db.gift_codes

    def add_code(self, code, created_for=None):
        """
        Adds a gift code to the mongodb collection
        """
        entry = {
            "code": code,
            "created_at": datetime.datetime.utcnow(),
            "used_at": None,
            "created_for": created_for,
        }
        if self.table.find_one({"code": code}):
            raise ValueError(
                "This gift code already exists. Pick another one.")
        self.table.insert_one(entry)

    def get_gift_code_count_by_user(self, username):
        """
        Return a the total number of gift codes created by a specific
        STEEM account.
        """
        return self.table.count({"created_for": username})

    def get_gift_codes_by_user(self, username):
        """
        Return a the gift codes created by a specificSTEEM account.
        """
        return self.table.find({"created_for": username})

    def get_code(self, code):
        """
        A helper function to get the related gift code entry from the database
        """
        return self.table.find_one({"code": code})

    def mark_code_as_used(self, code):
        """
        Updates the gift code in the database as used.
        """
        self.table.update_one(
            {"code": code},
            {'$set': {"used_at": datetime.datetime.utcnow()}})


class Infestor:

    def _prepare_claim_account_operation(self):
        return Operation(
            'claim_account',
            {
                'creator': self.creator,
                'fee': '0.000 STEEM',
                'extensions': [],
            }
        )

    def _prepare_create_claimed_account_operation(self, new_account_name, keys):
        op_data = {
            "creator": self.creator,
            "new_account_name": new_account_name,
            "json_metadata": "",
            "extensions": [],
        }
        for key_type, key_value in keys.items():
            if key_type == "master":
                continue
            if key_type == "memo":
                op_data.update({
                    "memo_key": key_value["public"]
                })
            else:
                op_data.update({
                    key_type: {
                        "account_auths": [],
                        "weight_threshold": 1,
                        "key_auths": [[key_value["public"], 1]],
                        "address_auths": []
                    }
                })

        op = Operation(
            'create_claimed_account',
            op_data,
        )
        return op

    def _get_active_key(self):
        if os.getenv('INFESTOR_ACTIVE_KEY'):
            return os.getenv('INFESTOR_ACTIVE_KEY')
        return getpass.getpass("Creator account's active key:\n")

    def _get_new_account_master_key(self):
        return getpass.getpass("Master key for the new account:\n")

    def __init__(self, creator, mongodb_connection_uri=None):
        self.creator = creator
        self.client = self.get_client_instance()
        if mongodb_connection_uri:
            self.gift_code_manager = GiftCodeManager(mongodb_connection_uri)
        else:
            self.gift_code_manager = GiftCodeManager(
                os.getenv("INFESTOR_MONGO_URI", "localhost"))

    def get_client_instance(self, keys=None):
        return Client(keys=keys)

    def claim_account(self, minimum_rc=None):
        """
        Broadcasts a claim_account operation.
        Creator account's RC and the cost of claim_account's RC is calculated
        beforehand.
        :return:
        """
        print(f"Fetching RC details of {self.creator}")

        creator_account = self.client.account(self.creator)
        pending_claimed_accounts = creator_account.raw_data \
            ["pending_claimed_accounts"]

        rc_details = creator_account.get_resource_credit_info()
        current_mana = rc_details["current_mana"] / 1000000

        if minimum_rc and rc_details["current_mana_percent"] < minimum_rc:
            print(f"RC of {self.creator} is not enough for the "
                  f"--minimum-rc requirement. "
                  f"{round(rc_details['current_mana_percent'], 2)}",
                  file=sys.stderr)
            sys.exit(-1)

        cost_of_claim_account = self.client.rc().get_cost(
            self._prepare_claim_account_operation())

        print(f"{self.creator} has {pending_claimed_accounts} "
              f"pending claimed accounts. Claiming an account currently "
              f"requires {cost_of_claim_account}MM mana.")

        print(f"{self.creator} has {int(current_mana)}MM mana available and"
              f" can claim {int(current_mana / cost_of_claim_account)}"
              f" more accounts.")

        if cost_of_claim_account > current_mana:
            print("Stopped. Insufficient mana.", file=sys.stderr)
            sys.exit(-1)

        active_key = self._get_active_key()

        self.client.keys = [active_key, ]
        self.client.broadcast(self._prepare_claim_account_operation())

        pending_claimed_accounts = self.client.account(
            self.creator).raw_data["pending_claimed_accounts"]

        print(f"Success! {self.creator} now has {pending_claimed_accounts} "
              f"pending claimed accounts.")

    def create_claimed_account(self, new_account_name):
        """
        Creates a discounted claimed account.
        Corresponding private and public keys are generated based on the
        master key.
        """

        if not new_account_name:
            print(
                "Add a --new-account-name <account_name> to the command.",
                file=sys.stderr
            )
            sys.exit(-1)

        active_key = self._get_active_key()
        new_account_master_key = self._get_new_account_master_key()

        # check the username is already exists?
        accounts = self.client.get_accounts([new_account_name])
        if len(accounts):
            print(
                f"{new_account_name} already exists. Choose another username.",
                file=sys.stderr)
            sys.exit(-1)

        keys = generate_keys(new_account_name, new_account_master_key)
        op = self._prepare_create_claimed_account_operation(
            new_account_name, keys)

        creator_account = self.client.account(self.creator)
        rc_details = creator_account.get_resource_credit_info()
        pending_claimed_accounts = creator_account.raw_data \
            ["pending_claimed_accounts"]

        if pending_claimed_accounts == 0:
            print(
                f"Stopped. {self.creator} has 0 pending claimed accounts.",
                file=sys.stderr
            )
            sys.exit(-1)

        current_mana = rc_details["current_mana"] / 1000000
        cost_of_create_account = self.client.rc().get_cost(op)

        if cost_of_create_account > current_mana:
            print("Stopped. Insufficient mana.", file=sys.stderr)
            sys.exit(-1)

        self.client.keys = [active_key, ]
        self.client.broadcast(op)

        print(f"{new_account_name} is successfully created."
              f" Save your passwords:")
        for key, subkeys in keys.items():
            if 'public' in subkeys:
                print(f"Type: {key}")
                print(f"Public: {subkeys['public']}\nPrivate:"
                      f" {subkeys['private']}\n--")
            else:
                print(f"Master Password: {subkeys}\n--")

    def add_gift_code(self, code):
        if not code:
            print("You need to pass --code as an argument.",
                  file=sys.stderr)
            sys.exit(-1)
        self.gift_code_manager.add_code(code)
        print(f"Gift code: {code} is added to the database.")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "action",
        help="An action to perform.",
        choices=["claim_account", "create_claimed_account", "add_gift_code"])
    parser.add_argument(
        '--creator', help="Creator account")
    parser.add_argument('--new-account-name', help="New account name")
    parser.add_argument('--minimum-rc', help="Minimum RC in percent", type=int)
    parser.add_argument('--code', help="Gift Code")

    args = parser.parse_args()

    infestor = Infestor(args.creator)
    if args.action == "claim_account":
        infestor.claim_account(minimum_rc=args.minimum_rc)
    elif args.action == "create_claimed_account":
        infestor.create_claimed_account(args.new_account_name)
    elif args.action == "add_gift_code":
        infestor.add_gift_code(args.code)


if __name__ == '__main__':
    main()



import os
import random

from flask import Flask, render_template, request, g
from lightsteem.client import Client as LightSteemClient
from lightsteem.helpers.account import Account
from steemconnect.client import Client as ScClient
from .main import Infestor
from .utils import username_is_valid, generate_random_password, generate_keys

app = Flask(__name__)
app.debug = True
INFESTOR_CREATOR_ACCOUNT = os.getenv('INFESTOR_CREATOR_ACCOUNT')
INFESTOR_ACTIVE_KEY = os.getenv('INFESTOR_ACTIVE_KEY')
INFESTOR_MONGO_URI = os.getenv("INFESTOR_MONGO_URI", "localhost")
SC_CLIENT_ID = os.getenv("INFESTOR_SC_CLIENT_ID", "infestor.app")
SC_SECRET = os.getenv("INFESTOR_SC_SECRET")
SITE_URL = os.getenv("INFESTOR_SITE_URL", "http://localhost:8000")
MINIMUM_REP = os.getenv("INFESTOR_MINIMUM_REP", "60")
OPERATOR_WITNESS = os.getenv("INFESTOR_OPERATOR_WITNESS", "emrebeyler")


@app.before_request
def set_lightsteem_client():
    g.lightsteem_client = LightSteemClient()
    g.infestor = Infestor(
        INFESTOR_CREATOR_ACCOUNT,
        mongodb_connection_uri=INFESTOR_MONGO_URI)
    g.operator_witness = OPERATOR_WITNESS
    g.minimum_rep = int(MINIMUM_REP)
    if os.getenv('INFESTOR_FOOTER_TEMPLATE'):
        with open(os.getenv('INFESTOR_FOOTER_TEMPLATE'), 'r') as f:
            g.footer = f.read()

    if not INFESTOR_ACTIVE_KEY or not INFESTOR_CREATOR_ACCOUNT:
        raise RuntimeError("Missing environment variables: "
                           "INFESTOR_CREATOR_ACCOUNT, INFESTOR_ACTIVE_KEY ")


@app.route('/', methods=["GET", "POST"])
def index():
    creator_account = g.lightsteem_client.get_accounts(
        [INFESTOR_CREATOR_ACCOUNT])[0]

    if creator_account["pending_claimed_accounts"] == 0:
        return "Free account pool is exhausted. Please, try again later on."

    gift_code = request.form.get(
        "gift_code", request.args.get("gift_code", ""))
    username = request.form.get("username", "")

    defaults = {
        "gift_code": gift_code,
        "username": username,
    }

    if request.method == 'GET':
        if gift_code:
            # validate the gift code from internal mongo database
            if not g.infestor.gift_code_manager.code_is_valid(gift_code):
                defaults.update({"error": "Invalid gift code."})
                return render_template(
                    "create_account.html",
                    **defaults
                )
        return render_template(
            "create_account.html",
            **defaults
        )
    else:

        # validate the gift code from internal mongo database
        if not g.infestor.gift_code_manager.code_is_valid(gift_code):
            defaults.update({"error": "Invalid gift code."})
            return render_template(
                "create_account.html",
                **defaults
            )

        # check the username is a valid username
        if not username_is_valid(defaults.get("username")):
            error = "Invalid username"
            defaults.update({"error": error})
            return render_template(
                "create_account.html",
                **defaults
            )

        # check the username is already taken
        if len(g.lightsteem_client.get_accounts([defaults.get("username")])):
            error = "Username is already taken."
            defaults.update({"error": error})
            return render_template(
                "create_account.html",
                **defaults
            )

        # generate the random password and keys
        password = generate_random_password()
        keys = generate_keys(username, password, exclude_master=True)

        # attach active key of the creator to the lightsteem client
        g.lightsteem_client.keys = [INFESTOR_ACTIVE_KEY, ]

        op = g.infestor._prepare_create_claimed_account_operation(
            username,
            keys,
        )

        defaults.update({
            "password": password,
            "keys": keys,
        })

        try:
            g.lightsteem_client.broadcast(op)
        except Exception as e:
            print(e)
            error = "roadcasting the create_claimed_account transaction"
            defaults.update({
                "error": error,
            })
            return render_template(
                "create_account.html",
                **defaults
            )

        g.infestor.gift_code_manager.mark_code_as_used(gift_code)
        return render_template("success.html", **defaults)


@app.route('/login', methods=["GET"])
def login():
    sc_client = ScClient(client_id=SC_CLIENT_ID, client_secret=SC_SECRET)
    login_url = sc_client.get_login_url(
        f"{SITE_URL}/gift-codes",
        "login",
    )
    return render_template("login.html", login_url=login_url)


@app.route('/gift-codes/', methods=["GET"])
def gift_codes():
    sc_client = ScClient(access_token=request.args.get("access_token"))
    me = sc_client.me()
    if 'error' in me:
        return "Invalid access token"

    # if the user already claimed their gift codes,
    # then there is no need to create new gift codes
    if g.infestor.gift_code_manager.get_gift_code_count_by_user(
            me["account"]["name"]) > 0:
        gift_codes = g.infestor.gift_code_manager.get_gift_codes_by_user(
            me["account"]["name"]
        )
        return render_template("gift_codes.html", user=me,
                               gift_codes=gift_codes)

    # quick hack
    # no need to fill all account data into Lightsteem.helpers.Account
    # since all we're interested in is reputation figure.
    acc = Account(client=g.lightsteem_client)
    acc.raw_data = {"reputation": me["account"]["reputation"]}

    gift_code_count = 0
    if acc.reputation() < g.minimum_rep:
        error = "Your reputation is not enough to claim a free account."
        return render_template("gift_codes.html", error=error)
    else:
        gift_code_count += 1
        # check if the account is eligible for the bonus
        if OPERATOR_WITNESS in me["account"]["witness_votes"]:
            gift_code_count += 1

    # create gift_codes based on the *gift_code_count*
    for i in range(gift_code_count):
        code = random.randint(1000000, 999999999)
        g.infestor.gift_code_manager.add_code(
            code, created_for=me["account"]["name"])
    gift_codes = g.infestor.gift_code_manager.get_gift_codes_by_user(
        me["account"]["name"])

    return render_template("gift_codes.html", user=me, gift_codes=gift_codes)



import random
import re
import string

from lightsteem.broadcast.key_objects import PasswordKey


def generate_keys(new_account_name, master_key, exclude_master=False):
    keys = {
        "master": master_key,
    }
    if exclude_master:
        keys = {}
    for key_type in ['posting', 'active', 'owner', 'memo']:
        private_key = PasswordKey(
            new_account_name, master_key, key_type
        ).get_private_key()
        keys[key_type] = {
            "public": str(private_key.pubkey),
            "private": str(private_key),
        }
    return keys


def username_is_valid(username):
    # regex pattern is found at
    # https://steemit.com/programming/@cryptosharon/the-5-rules-of-a-valid-username-on-the-steem-blockchain-and-a-3-sbd-contest-to-make-an-account-name-validation-regex#@cryptosharon/re-eonwarped-re-cryptosharon-re-eonwarped-re-cryptosharon-re-eonwarped-re-cryptosharon-re-artopium-re-cryptosharon-the-5-rules-of-a-valid-username-on-the-steem-blockchain-and-a-3-sbd-contest-to-make-an-account-name-validation-regex-20180313t214044982z
    return bool(re.search('^[a-z](-[a-z0-9](-[a-z0-9])*)?(-[a-z0-9]|[a-z0-9])'
                          '*(?:\.[a-z](-[a-z0-9](-[a-z0-9])*)?(-[a-z0-9]|'
                          '[a-z0-9])*)*$', username))


def generate_random_password():
    return 'P5' + ''.join(
        random.choices(
            string.ascii_uppercase + string.digits + string.ascii_lowercase,
            k=50))



from setuptools import setup, find_packages

setup(
    name='infestor',
    version='0.1.2',
    packages=find_packages(),
    url='http://github.com/emre/transmitter',
    license='MIT',
    author='emre yilmaz',
    author_email='mail@emreyilmaz.me',
    description='CLI app to claim and create discounted accounts'
                ' on the STEEM blockchain.',
    entry_points={
        'console_scripts': [
            'infestor = infestor.main:main',
        ],
    },
    include_package_data=True,
    zip_safe=False,
    install_requires=["lightsteem", "pymongo", "flask", "gunicorn", "steemconnect"],
)
